import numpy as np
import scipy.stats.mstats as scistats
import sklearn.metrics as scimet
import sklearn.preprocessing as skpp
from fastnet.log import logger


def sp(truePositive, trueNegative):
    """function ret  = calcSP(sp(truePositive, trueNegative))
    Calculates the normalized [0,1] SP value.
        Input Parameters are:
            truePositive -> The detection efficiency as an scalar or a (N,) array.
            trueNegative -> The true negative (1 - false alarm) as an scalar or a (N,) array.

    THIS FUNCTION WORKS ONLY FOR A 2-CLASS CLASSIFIER.
    """
    m = np.concatenate((np.atleast_2d(truePositive), np.atleast_2d(trueNegative)))
    ret = np.sqrt(scistats.gmean(m, axis = 0) * np.mean(m, axis = 0))
    return ret[0] if ret.shape[0] == 1 else ret


def roc(outSignal, outNoise, numPts = 1000, algo = 'scikit'):
    """function [det, fa, sp, ppv, ppn, acu, thres] = roc(outSignal, outNoise, numPts = 1000, algo = 'scikit')
    Returns the Receiver Operational Characterists.
    Input Parameters are:
        outSignal                  -> The output (as (N,) numpy vector) generated by your classifier when signal-related data is applied to it.
        outNoise                   -> The output (as (N,) numpy vector) generated by your classifier when noise-related data is applied to it.
        numPts                     -> How many points the ROC should have (defaults to 1000). Ignored if algo != 'matlab'.
        algo ('matlab', 'scikit')  -> Can be either the Matlab algo or use the SciKit-Learn roc function (default).
    Returns:
        det   -> Detection efficiency (true positive probability) per decision threshold as a numpy (N,) vector).
        fa    -> False alarm efficiency (false positive probability) per decision threshold as a numpy (N,) vector).
        sp    -> The SP value per decision threshold as a numpy (N,) vector).
        ppv   -> The Positive Predicted Value per decision threshold as a numpy (N,) vector).
        npv   -> The Negative Predicted Value per decision threshold as a numpy (N,) vector).
        acu   -> Acuity per decision threshold as a numpy (N,) vector).
        thres -> The decision threshold as a numpy (N,) vector)

    THIS FUNCTION WORKS ONLY FOR A 2-CLASS CLASSIFIER.
    """

    def matlab_roc(outSignal, outNoise, numPts):
        fullData = np.concatenate((outSignal, outNoise))
        doNorm = (fullData.min() < -1) or (fullData.max() > 1)
        if doNorm:
            logger.info('Normalizing output values so they can be within the [-1,1] range.')
            pp = skpp.MinMaxScaler(feature_range=(-1, 1))
            pp.fit(fullData.reshape(-1,1))
            outSignal = pp.transform(outSignal.reshape(-1,1)).flatten()
            outNoise = pp.transform(outNoise.reshape(-1,1)).flatten()
        cutVec = np.linspace(-1,1,numPts)
        det = np.zeros(numPts)
        fa = np.zeros(numPts)
        for i, cut in enumerate(cutVec):
            det[i], fa[i] = efficiency(outSignal, outNoise, cut)[0:2]
        if doNorm: cutVec = pp.inverse_transform(cutVec.reshape(-1,1)).flatten()
        return fa, det, cutVec

    def sci_roc(outSignal, outNoise, *kwargs):
        #Converting to scipy format (output and target)
        output = np.concatenate((outSignal, outNoise))
        target = np.concatenate( (np.ones(outSignal.shape), -np.ones(outNoise.shape)) )
        fa, det, thres = scimet.roc_curve(target, output, pos_label = 1)
        return fa, det, thres

    assert algo in ['scikit', 'matlab'] , 'Algoritmo de cálculod e ROC inválido!'
    roc_algo = matlab_roc if algo == 'matlab' else sci_roc
    fa, det, thres = roc_algo(outSignal, outNoise, numPts)

    nSignal = len(outSignal)
    nNoise = len(outNoise)
    tn = 1-fa
    fn = 1-det
    vpp = det / (det + fa)
    vpn = tn / (tn + fn)
    acc = (det*nSignal + tn*nNoise) / (nSignal + nNoise)
    spVec = sp(det, tn)

    return det, fa, spVec, vpp, vpn, acc, thres



def efficiency(outSignal, outNoise, cut = 0.):
    """function [det, fa, sp, ppv, ppn, acu] = efficiency(outSignal, outNoise, cut)
    Returns the efficiency parameters for a detecor.
    Input Parameters are:
        outSignal -> The output (as (N,) numpy vector) generated by your classifier when signal-related data is applied to it.
        outNoise  -> The output (as (N,) numpy vector) generated by your classifier when noise-related data is applied to it.
        cut       -> The decision threshold (defaults to 0). The function considers as signal any output >= cut.
    Returns:
        det -> Detection efficiency (true positive probability)
        fa  -> False alarm efficiency (false positive probability)
        sp  -> The SP value
        ppv -> The Positive Predicted Value
        npv -> The Negative Predicted Value
        acu -> Acuity

    THIS FUNCTION WORKS ONLY FOR A 2-CLASS CLASSIFIER.
    """

    nSignal = len(outSignal)
    nNoise = len(outNoise)
    det = np.sum(outSignal >= cut) / nSignal
    fa = np.sum(outNoise >= cut) / nNoise
    tn = 1-fa
    fn = 1-det
    acu = (det*nSignal + tn*nNoise) / (nSignal + nNoise)
    vpp = det / (det + fa)
    vpn = tn / (tn + fn)
    spVal = sp(det, tn)
    return det, fa, spVal, vpp, vpn, acu



#This module contains classes that provides a high level interface for data loading.
import torch
import torch.utils.data as dm
import numpy as np
from fastnet.log import logger
import sklearn.model_selection as sms

def to_tensor(mat):
    '''Converts the passed numpy.array to a torch tensor'''
    return torch.tensor(mat, dtype = torch.float32)


def get_classes_input_data(data, classLabels = [+1, -1]):
    """Returns the input data corresponding to each class label.
       Parameters:
           - data: a (Input, Target) tuple where Input is the mixed input
                   data and Target their corresponding target.
           - classLabels: a list of class labels (default to [+1, -1])
       Returns: a tuple with the Input data for each class label in the same
                order as specified in classLabels.

       THIS WORKS FOR 2-CLASS CASE ONLY"""
    inData, targData = data
    targData = targData.flatten()
    return tuple([inData[targData == l,:] for l in classLabels])


def get_num_events_per_class(targList):
    """For a taget vector, tells how many events each class has.
    Returns a tuple with 2 lists ([labes], [counts]).The first provides each
    unique class label, and the second how many events each class has.

    THIS WORKS FOR 2-CLASS CASE ONLY"""
    labels, counts = np.unique(targList, return_counts = True)
    return labels, counts


def get_train_sampler(target):
    """
    Creates a weighted random sampler. The data sampler returned by this function
    will have the following features:
       - It will draw samples among classes with equal probability, regardless of number of events inbalances among classes.
       - Samples will be randomly selected.

    Input:
      - target: a [N x 1] numpy array containing all targets of all events of all classes.

    Return:
      - the weighted random sampler created after analysing the passed target vector.
    """
    #Getting the number of events per class.
    labels, counts = get_num_events_per_class(target)
    N = target.shape[0]
    logger.debug('Number of (unbalanced) events per target symbol {}: {}'.format(labels, counts))
    logger.debug('Total set size: {}'.format(N))

    largestClassSize = counts.max()
    weights = np.zeros(N)
    for label, numEv in zip(labels, counts):
        wVal = largestClassSize / numEv
        weights[target[:,0] == label] = wVal
        logger.debug('Weight for class {}: {}'.format(label, wVal))

    numSamples2Draw = int(largestClassSize * len(counts))
    logger.debug('Total available samples to retrieve at each epoch: {}'.format(numSamples2Draw))

    return dm.WeightedRandomSampler(weights, numSamples2Draw, replacement = True)



def get_datasets(dataList, test_size = 0.5):
    """
    Will get a list of class data and class labels and create proper training and testing
    sets.

    Input:
        dataList: list of tuples. Each tuple represents a class and should contain
                    the input data (a numpy array where each row is an event)
                    and label (a scalar) for a class. Example: [(X, +1), (Y,-1)].
        test_size: the percentual size [0,1] of the testing set.

    Output:
      - train: a (input, target) training tuple where input is a numpy array containing the input event where each event is a row.
               Target is a numpy array containing the target of each corresponding input (each target is a row).
      - test: a (input, target) testing tuple where input is a numpy array containing the input events where each event is a row.
               Target is a numpy array containing the target of each corresponding input (each target is a row).
    """

    inTrnList = []
    targTrnList = []
    inTstList = []
    targTstList = []

    classIdx = 1
    for input, label in dataList:
        nSamples = input.shape[0]

        #Creating a target vector using the scalar label passed.
        target = np.zeros([nSamples,1]) + label

        #Splitting the class data into training and testing sets.
        inTrn, inTst, targTrn, targTst = sms.train_test_split(input, target, test_size = test_size)
        logger.debug('Training / testing division for class {}: {}/{}'.format(classIdx, inTrn.shape[0], inTst.shape[0]))
        inTrnList.append(inTrn)
        targTrnList.append(targTrn)
        inTstList.append(inTst)
        targTstList.append(targTst)

        classIdx += 1

    #Concatenating everything to create a single dataset.
    inTrnList = np.concatenate(inTrnList)
    targTrnList = np.concatenate(targTrnList)
    inTstList = np.concatenate(inTstList)
    targTstList = np.concatenate(targTstList)

    #Creating the dataset (input x target) for the training / testing sets
    train = (inTrnList, targTrnList)
    test = (inTstList, targTstList)

    return train, test


def get_balanced_batched_loader(train, test, batch_size = 0):
    """
    Generates the train, test data loaders. The train dataloader will
    resolve any class number o samples inbalance by adjusting the samples
    drawing probability so the loader will draw samples from all classes with
    equal probability. In addition, the training loader will always ramdomly
    select samples. The test loader returned does NOT treat classes inbalances
    and always returns events in the same order.

    Input:
      - train: a (input, target) training tuple where input is a numpy array containing the input events where each event is a row.
               Target is a numpy array containing the target of each corresponding input (each target is a row).
      - test: a (input, target) testing tuple where input is a numpy array containing the input events where each event is a row.
               Target is a numpy array containing the target of each corresponding input (each target is a row).
      - batch_size: the training batch size to be used. Set it to 0 (zero) if you want to use the entire training set at each epoch. Defaults to zero.

    Returns:
      - trainDataLoader: the data loader for the trainign set already in torch.tensor format (ready to be used in the training).
      - testDataLoader: the data loader for the testing set already in torch.tensor format (ready to be used in the training).
    """

    #We will create a sampler that takes into considerantion
    #class inbalances regarding number of samples.
    trainSampler = get_train_sampler(train[1])

    if batch_size == 0: batch_size = trainSampler.num_samples
    logger.debug('Training batch size per epoch: {}'.format(batch_size))

    #Placing the data finally into a torch.tensor right before the training.
    train = dm.TensorDataset(to_tensor(train[0]), to_tensor(train[1]))
    test = dm.TensorDataset(to_tensor(test[0]), to_tensor(test[1]))

    trainDataLoader = dm.DataLoader(train, sampler = trainSampler, batch_size = batch_size)
    testDataLoader = dm.DataLoader(test, batch_size = len(test))
    logger.debug('Total testing events to be employed each epoch: {}'.format(len(test)))

    return trainDataLoader, testDataLoader



class CrossValidationManager:
    """
    This class manages cross-validation training by making sure
    data sets are split in training / test sets according to a cross validation
    object passed.
    """
    def __init__(self, crossValObj, dataList):
        """
        Input:
          - crossValObj: of the the many cross validation functions available in sklearn.model_selection.
                         the object passed will define how the training / testing sets will be drawn (Jack-knife, leave-one-out, etc.).
          - dataList:    list of tuples. Each tuple represents a class and should contain
                         the input data (a numpy array where each row is an event)
                         and label (a scalar) for a class. Example: [(X, +1), (Y,-1)].
        """
        inputList = []
        targetList = []
        classLabels = []
        for input, label in dataList:
            #Creating a target vector using the scalar label passed.
            inputList.append(input)
            targetList.append(np.zeros([input.shape[0],1]) + label)

        self.input = np.concatenate(inputList)
        self.target = np.concatenate(targetList)
        classLabels = self.target.flatten()
        self.Draws = [(trn, tst) for trn, tst in crossValObj.split(classLabels, classLabels)]


    def split(self, batch_size = 0, returnCounter = False):
        """
        Draw a new training / testing set to be used for the training.

        Input:
          - batch_size: the training batch size to be used. Set it to 0 (zero)
                        if you want to use the entire training set at each epoch. Defaults to zero.
          - returnCounter: if True, along with the training, testing sets, it will return also
                           the index of the set being draw (useful for control by the calling function).

        Return:
          A list where each element is a tuple of the format (k, train, test) if returnCounter = True
          or (train, test) if returnCounter = False, where:

            - k: is the index of the set being drawn
            - train: a (input, target) training tuple where input is a numpy array containing the input events where each event is a row.
                     Target is a numpy array containing the target of each corresponding input (each target is a row).
            - test: a (input, target) testing tuple where input is a numpy array containing the input events where each event is a row.
                    Target is a numpy array containing the target of each corresponding input (each target is a row).
        """
        ret = []
        for i, d in enumerate(self.Draws):
            v = []
            if returnCounter: v.append(i)
            trn, tst = d
            logger.debug('Deal "{}" - training indices: {}'.format(i, trn))
            logger.debug('Deal "{}" - testing indices: {}'.format(i, tst))
            v.append( (self.input[trn,:], self.target[trn,:]) )
            v.append( (self.input[tst,:], self.target[tst,:]) )
            ret.append(tuple(v))
        return ret



#!/usr/bin/env python3

#Falta desenvolver:
#    1) Validação cruzada
#    2) Fazer a classe de treinamento trabalhar com classes (estilo seu cell array do matlab)

import os
import time
os.environ['FASTNET_LOG_LEVEL'] = 'ERROR'

import sys
import torch.nn
import plotly.offline as pyo
import plotly.graph_objs as go
import numpy as np
import fastnet.train as fn
import fastnet.metrics as met
import fastnet.data_manager as dm


def plot_output(data, model, title):
    hist = []
    for i,c in enumerate(data):
        className = 'Class {}'.format(i+1)
        classOutData = model(c[0])
        hist.append(go.Histogram(x = classOutData, name = className, opacity=0.75))

    layout = go.Layout(title = title,
                        xaxis =  dict(title ='Neural Network Output'),
                        yaxis =  dict(title = 'Count'),
                        hovermode= 'closest')

    fig = go.Figure(data = hist, layout = layout)
    pyo.plot(fig, filename= '/tmp/' + title + '.html')


def plot_input(c1,c2):
    pc1 = go.Scatter(x = c1[:,0],
                      y = c1[:,1],
                      mode = 'markers',
                      name = 'Class 1')
    pc2 = go.Scatter(x = c2[:,0],
                      y = c2[:,1],
                      mode = 'markers',
                      name = 'Class 2')

    layout = go.Layout(title = "Patterns Original Distribution",
                        xaxis =  dict(title ='Var 1'),
                        yaxis =  dict(title = 'Var 2'),
                        hovermode= 'closest')

    fig = go.Figure(data = [pc1, pc2], layout = layout)
    pyo.plot(fig, filename='/tmp/input.html')



def plot_roc(det, fa, tit):
    det = np.atleast_2d(det)
    fa = np.atleast_2d(fa)
    pc1 = [go.Scatter(x = 100*fa[i,:],
                      y = 100*det[i,:],
                      mode = 'lines',
                      name = 'ROC') for i in range(det.shape[0])]
    layout = go.Layout(title = "ROC ({})".format(tit),
                        xaxis =  dict(title ='False Alarm'),
                        yaxis =  dict(title = 'Detection Efficiency'),
                        hovermode= 'closest')

    fig = go.Figure(data = pc1, layout = layout)
    pyo.plot(fig, filename= '/tmp/' + tit + '-ROC.html')


def plot_sp(spVec):
    sp = 100*spVec
    pc1 = [go.Scatter(x = [1],
                      y = [sp.mean()],
                      error_y = dict(type='data', array=[sp.std()], visible=True),
                      mode = 'lines+markers',
                      name = 'Max SP.')]
    layout = go.Layout(title = "Max SP Analysis",
                        xaxis =  dict(title ='Classifier'),
                        yaxis =  dict(title = 'SP (x100)'),
                        hovermode= 'closest')

    fig = go.Figure(data = pc1, layout = layout)
    pyo.plot(fig, filename= '/tmp/Max SP.html')



def plot_errors(trn, val):
    plot = []
    for i in range(trn.shape[0]):
        plot.append(go.Scatter(y = trn[i,:],
                                    mode = 'lines+markers',
                                    name = 'Training Error',
                                    line = dict(color = 'blue')))
        plot.append(go.Scatter(y = val[i,:],
                                    mode = 'lines+markers',
                                    name = 'Validation Error',
                                    line = dict(color = 'red')))

    layout = go.Layout(title = "Training Evolution",
                        xaxis =  dict(title ='Epoch'),
                        yaxis =  dict(title = 'MSE'),
                        hovermode= 'closest')

    fig = go.Figure(data = plot, layout = layout)
    pyo.plot(fig, filename='/tmp/Training Evolution.html')


def get_dataset(nData):
    #Generating data set.
    c1 = np.random.randn(nData, 2) + 3
    c2 = np.random.randn(nData, 2)
    plot_input(c1,c2)
    return [(c1, +1), (c2, -1)]



#Global level parameters
nData = 1000
batchSize = 0
hidNodes = 3
epochs = 100
evalFunc = 'sp'
nTrains = 2
saveData = True
nDeals = 5
numROC = 1000
testSize = 0.5
trfFunction = [torch.nn.Tanh, torch.nn.Tanh]

#Creating dataset.
data = get_dataset(nData)

#Resultados para a rede nao treinada.
net = fn.BinaryClassificationNetwork([2, hidNodes, 1], trfFunction, evalFunc = evalFunc)
outSignal = net(data[0][0])
outNoise = net(data[1][0])
det, fa, _, _, _, _, _ = met.roc(outSignal, outNoise, numPts = numROC, algo = 'matlab')
plot_output(data, net, 'Saída Antes do Treino')
plot_roc(det, fa, 'Antes do Treino')


ret = None
start = time.time()
if nDeals == 1:
    print('Using single-training')
    ret = fn.single_training(data, [hidNodes], trfFunction,
                                testSize = testSize, evalFunc = evalFunc,
                                nTrains = nTrains, batchSize = batchSize,
                                epochs = epochs, numROC = numROC)
else:
    print('Using cross-validation trainning')
    ret = fn.cross_validation_training(data, [hidNodes], trfFunction,
                                        testSize = testSize, evalFunc = evalFunc,
                                        nTrains = nTrains, batchSize = batchSize,
                                        epochs = epochs, numROC = numROC, nDeals = nDeals)

end = time.time()

print('Training elapsed time: {0:0.2f} seconds.'.format(end-start))

#Getting best network.
idx = ret.maxSP.argmax()
net = ret.net[idx]
tstDataAux = dm.get_classes_input_data(ret.tstData[idx])
tstData = [(tstDataAux[0], +1), (tstDataAux[1], -1)]


#Resultados pos treino
plot_output(tstData, net, 'Saída Após o Treino')
plot_roc(ret.det, ret.fa, 'Após o Treino')
plot_errors(ret.trnError, ret.tstError)
plot_sp(ret.maxSP)



from fastnet.log import logger
import fastnet.data_manager as dm
import fastnet.metrics as met
import numpy as np

def relevance(net, trn, val, mode = 'mse', cut = 0.):
    """r = relevance(net, trn, val, mode, cut)
    Performs relevance analysis.

    Inputs:
      - net: a regression/classification trained model. It must be an object that provides
             the model output by calling its __call__ method.
      - trn: the data used to train the model.
      - val: the dataset used to validate the model (as a (input, target) tuple where input and target are numpy.arrays).
      - mode [mse, sp]: tells whether to calculate the relevance using mse or sp (defaults to 'mse').
      - cut: cut is the cut threshold between classes (to be used only if mode = 'sp'). Defaults to 0.

    The function returns a numpy vector with the output goal (MSE or SP) deviation for each removed input.

    WARNING: THIS FUNCTION WORKS FOR 2 CLASSES CASE ONLY
    """

    assert mode in ('mse', 'sp'), 'Mode must be either "sp" or "mse"'

    #Getting the mean value of each input variable.
    mdata = trn.mean(axis=0)
    if mode == 'mse':
        logger.info('Doing relevance analysis by MSE.')
        return do_by_mse(mdata, net, val[0])
    else:
        logger.info('Doing relevance analysis by SP.')
        return do_by_sp(mdata, net, val, cut)


def do_by_mse(mdata, net, val):
    nDim = val.shape[1]
    ret = np.zeros(nDim)
    outRef = net(val)

    for i in range(nDIM):
        #Creating a copy of the validation set and replacing the i-th input
        #by its mean value.
        aux = val.copy()
        aux[;,i] = mdata[i]

        out = net(aux)

        ret[i] = ((outRef - out)**2).mean()

    return ret


def do_by_sp(mdata, net, val, cut):
    nDim = val.shape[1]
    ret = np.zeros(nDim)
    signal, noise = dm.get_classes_input_data(val)
    _,_, spRef ,_,_,_ = met.efficiency(net(signal), net(noise), cut)

    for i in range(nDIM):
        #Creating a copy of the validation set and replacing the i-th input
        #by its mean value.
        auxSignal = signal.copy()
        auxNoise = noise.copy()
        auxSignal[;,i] = mdata[i]
        auxNoise[;,i] = mdata[i]
        _,_, sp ,_,_,_ = met.efficiency(net(auxSignal), net(auxNoise), cut)
        ret[i] = spRef - sp



import torch
import torch.nn
import torch.optim
import numpy as np
import copy
import collections
import math
from fastnet.log import logger
import fastnet.data_manager as dm
import fastnet.utils as ut
import fastnet.metrics as met
from sklearn.model_selection import StratifiedShuffleSplit


class SPLoss(torch.nn.modules.loss._Loss):
    """Implements SP loss function according to PyTorch requirements for loss functions"""

    def __init__(self, num_roc_pts = 200, signal_label = 1., noise_label = -1.):
        """
        - num_roc_pts: scalar definign how many points to be used when evaluating the ROC. Defaults to 200.
        - signal_label: a scalar defining the label to be used for the signal class. Defaults to +1.
        - noise_label: a scalar defining the label to be used for the noise class. Defaults to -1.
        """
        super(SPLoss, self).__init__()
        self.num_roc_pts = num_roc_pts
        self.signal_label = signal_label
        self.noise_label = noise_label

    def forward(self, output, target):
        """
        Computes the SP loss based on the output x target arrays passed.
        Returns the maximum SP found for the generated ROC.
        WORKS ONLY FOR BINARY CLASSIFICATION NETWORKS.
        """
        cutVec = np.linspace(self.noise_label, self.signal_label, self.num_roc_pts)
        signal = output[target == self.signal_label]
        noise = output[target == self.noise_label]
        nSignal = float(len(signal))
        nNoise = float(len(noise))
        maxSP = -torch.ones(1, dtype = torch.float32)
        efics = torch.zeros(2, dtype = torch.float32)
        for cut in cutVec:
            efics[0] = torch.sum(signal >= cut, dtype = torch.float32) / nSignal #Detection efficiency
            efics[1] = torch.sum(noise < cut, dtype = torch.float32) / nNoise # 1 - false alarm
            sp = torch.sqrt( efics.prod().sqrt() * efics.mean() )
            maxSP = torch.max(maxSP, sp)
        return maxSP


class BinaryClassificationNetwork:
    """Class to implement the training of a binary classifier."""

    def __init__(self, numNodes, trfFunc, evalFunc = 'mse'):
        """
        - numNodes: a list of scalars defining the number of nodes in each layer (including the input layer).
        - trfFunc: a list of torch.nn available transfer functions for all hidden layer and the output layer.
        - evalFunc: a string defining the desired evaluation function. Values can be 'mse' or 'sp'. Defaults to 'mse'.
        """
        assert evalFunc in ['mse', 'sp'], 'Invalid evaluation function for validation step.'
        assert (len(numNodes)-1) == len(trfFunc), 'numNodes and trFunc vectors length do not make sense! trFunc must have 1 element less than numNodes.'
        topology = []
        for i, func in enumerate(trfFunc):
            topology.append(torch.nn.Linear(numNodes[i], numNodes[i+1]))
            topology.append(func())
        self.model = torch.nn.Sequential(*tuple(topology))
        self.loss_fn_trn = torch.nn.MSELoss(reduction='mean')
        if evalFunc == 'mse':
            self.loss_fn_val = torch.nn.MSELoss(reduction='mean')
            logger.info('Stop criteria: MSE')
            self.evalFix = 1.
        elif evalFunc == 'sp':
            self.loss_fn_val = SPLoss()
            logger.info('Stop criteria: SP')
            self.evalFix = -1.
        self.optimizer = torch.optim.Rprop(self.model.parameters())
        self.bestModel = copy.deepcopy(self.model)


    def initWeights(self, m):
        """Initializes the network weights."""
        if type(m) == torch.nn.Linear:
            torch.nn.init.kaiming_uniform_(m.weight, a = math.sqrt(5))
            if m.bias is not None:
                fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(m.weight)
                bound = 1 / math.sqrt(fan_in)
                torch.nn.init.uniform_(m.bias, -bound, bound)


    def train(self, trnData):
        """Trains the neural network for just one epoch."""
        avg = 0.
        N = 0.
        for input, target in trnData:
            self.model.train()
            out = self.model(input)
            loss = self.loss_fn_trn(out, target)
            avg += loss.item()
            N += 1
            loss.backward()
            self.optimizer.step()
            self.optimizer.zero_grad()
        return avg / N


    def validate(self, val):
        """Validates the neural network for just one epoch."""
        self.model.eval()
        with torch.no_grad():
            #We assume here that the val set is not batched so we call
            #the next function of the data loader iterator so we can get
            #the full size of the valdiation set.
            inVal, targVal = next(val.__iter__())
            return self.loss_fn_val(self.model(inVal), targVal).item()


    def initializeAndTrain(self, train, val, epochs):
        """
        Forces a new weights initialization and then performs the training.
        This is useful when doing multiple weights initialization to dodge local minimum.
        """
        trnError = []
        valError = []
        minError = 10000000000
        self.model.apply(self.initWeights)
        bestModel = None
        for e in range(epochs):
            trnError.append(self.train(train))
            valE = self.validate(val)
            valError.append(valE)
            valE *= self.evalFix
            if valE < minError:
                bestModel = copy.deepcopy(self.model)
                minError = valE
                logger.debug('Best net found in epoch {}: {}'.format(e, self.evalFix*valE))
        return np.array(trnError), np.array(valError), minError, bestModel


    def fit(self, train, val, epochs = 100, numInitializations = 10, batch_size = 0):
        """
        Fit the neural network using the training data passed.

        Input:
          - train: a (input, target) training tuple where input is a list where each
                   element is a numpy array containing the input events of the i-th class where each event is a row.
                   Target is a numpy array containing the target of each corresponding input (each target is a row).
          - test: a (input, target) testing tuple where input is a list where each
                  element is a numpy array containing the input events of the i-th class where each event is a row.
                  Target is a numpy array containing the target of each corresponding input (each target is a row).
          - epochs: for how many epochs to train the model (save the best is employed,
                    so do not fear setting a large value if desired). Defaults to 100.
          - numInitializations: how many times re-initialize the network and re-train it.
                    Useful to dodge local minimus. If > 1, then just the network
                    with the best result for the testing set is returned. Defaults to 10.
          - batch_size: the training batch size to be used. Set it to 0 (zero)
                        if you want to use the entire training set at each epoch. Defaults to zero.
        """
        #Creating the torch.tensors datasets for handling the network data feeding.
        train, val = dm.get_balanced_batched_loader(train, val, batch_size)
        gbMinError = gbTrnError = gbValError = 10000000000
        self.bestModel = copy.deepcopy(self.model)
        for i in range(numInitializations):
            trnError, valError, minError, bestModel = self.initializeAndTrain(train, val, epochs)
            logger.debug('Min error for initialization was {}'.format(minError))
            if minError < gbMinError:
                self.bestModel = bestModel
                gbTrnError = trnError
                gbValError = valError
                gbMinError = minError
        logger.debug('Best network min error = {}'.format(gbMinError))
        return gbTrnError, gbValError


    def propagate(self, vec):
        """
        Feedforward the input vector (numpy or torch tensor where each event is a row)
        through the network (trained or not).

        Returns a flattened [0 x N] numpy vector with the output obtained for each provided input.
        """
        return self.bestModel(vec)

    def __call__(self, vec):
        """
        Feedforward the input vector (numpy or torch tensor where each event is a row)
        through the network (trained or not).

        Returns a flattened [0 x N] numpy vector with the output obtained for each provided input.
        """
        if type(vec) is np.ndarray:
            logger.debug('Converting numpy matrix to torch.Tensor before propagating iot through the network.')
            vec = dm.to_tensor(vec)
        ret = self.bestModel(vec)
        return ret.detach().numpy().flatten()


def create_results_structure(nDeals, epochs, numROC):
    """
    Creates a structure to hold the network results

    Input:
      - nDeals: how many cross-validation deals to consider.
      - epochs: number of training epochs.
      - numROC: number of points to consider when generating the ROC.

    Returns a ObjectView object with the following attributes:
      - net: list for storing BinaryClassificationNetwork trained objects for each deal.
      - trnError: [nDeals, epochs] numpy.zeros matrix for storing training evolution errors.
      - tstError: [nDeals, epochs] numpy.zeros matrix for storing testing evolution scores.
      - det: [nDeals, numROC] numpy.zeros matrix for storing detection efficiency for each deal.
      - fa: [nDeals, numROC] numpy.zeros matrix for storing false alarme probability for each deal.
      - maxSP: [0, nDeals] numpy.zeros vector to store the maximum SP obtained for each deal.
      - thres: [0, nDeals] numpy.zeros vector to store the threshold where the maximum SP was obtained for each deal.
      - trnData: a list to hold the training data used for each deal.
      - tstData: a list to hold the testing data used for each deal.
    """
    ret = ut.ObjectView()
    ret.net = []
    ret.trnError = np.zeros([nDeals, epochs])
    ret.tstError = np.zeros([nDeals, epochs])
    ret.det = np.zeros([nDeals, numROC])
    ret.fa = np.zeros([nDeals, numROC])
    ret.maxSP = np.zeros(nDeals)
    ret.thres = np.zeros(nDeals)
    ret.trnData = []
    ret.tstData = []
    return ret


def save_results(ret, idx, trnError, tstError, net, trainData, tstData):
    """
    Saves the training results into the structure created by create_results_structure.

    Inputs:
      - ret: an ObjectView structure created by create_results_structure.
      - idx: the index of the training (deal) being performed.
      - trnError: [0, epochs] vector with the network training evolution errors.
      - tstError: [0, epochs] vector with the network testing evolution errors.
      - net: a BinaryClassificationNetwork trained objects.
      - trnData: training data used for the deal.
      - tstData: testing data used for the deal.

    Returns ret with its corresponding attributes populated for the idx-th deal.
    """

    numROC = ret.det.shape[1]
    inTst, targTst = tstData
    targTst = targTst.flatten()
    inSignal, inNoise = dm.get_classes_input_data(tstData)
    outSignal = net(inSignal)
    outNoise = net(inNoise)
    ret.det[idx,:], ret.fa[idx,:], spVec, _, _, _, thres = met.roc(outSignal, outNoise, numPts = numROC, algo = 'matlab')
    ret.net.append(net)
    ret.trnError[idx,:] = trnError
    ret.tstError[idx,:] = tstError
    ret.maxSP[idx] = spVec.max()
    ret.thres[idx] = thres[spVec.argmax()]
    ret.trnData.append(trainData)
    ret.tstData.append(tstData)
    return ret


def single_training(data, hiddenNodes, trfFunction, testSize = 0.5, evalFunc = 'sp', nTrains = 10, batchSize = 0, epochs = 100, numROC = 200):
    """
    single_training(data, hiddenNodes, trfFunction, testSize = 0.5, evalFunc = 'sp',
                    nTrains = 10, batchSize = 0, epochs = 100, numROC = 200)

    Performs a single training.

    Inputs:
      - data: list of tuples. Each tuple represents a class and should contain
              the input data (a numpy array where each row is an event)
              and label (a scalar) for a class. Example: [(X, +1), (Y,-1)].
      - hiddenNodes: a list stating how many nodes to be used in each hidden layer
                    (do not specify nodes for the input and output layers).
      - trfFunction: a list of torch.nn available transfer functions for all hidden layers and the output layer.
      - testSize: the percentual size [0,1] of the testing set. Defaults to 0.5.
      - evalFunc: the evalutiaon function to use ('mse' or 'sp'). Defaults to 'sp'.
      - nTrains: how many weights initialization to perform. Defaults to 10.
      - batchSize: epochs batch size. Defaults to zero (use all available data).
      - epochs: number of training epochs. Defaults to 100.
      - numROC: ROC size, when evaluating performance. Defaults to 200.

     Returns a populated ObjectView object created by create_results_structure.
    """

    train, test = dm.get_datasets(data, test_size = testSize)
    numNodes = [train[0].shape[1]] + hiddenNodes + [train[1].shape[1]]
    net = BinaryClassificationNetwork(numNodes, trfFunction, evalFunc = evalFunc)
    trnError, tstError = net.fit(train, test, epochs, numInitializations = nTrains, batch_size = batchSize)
    ret = create_results_structure(1, epochs, numROC)
    return save_results(ret, 0, trnError, tstError, net, train, test)


def cross_validation_training(data, hiddenNodes, trfFunction, testSize = 0.5, evalFunc = 'sp', nTrains = 10, batchSize = 0, epochs = 100, numROC = 200, nDeals = 20):
    """
    cross_validation_training(data, hiddenNodes, trfFunction, testSize = 0.5, evalFunc = 'sp',
                                nTrains = 10, batchSize = 0, epochs = 100, numROC = 200, nDeals = 20)

    Performs a cross-validation training.

    Inputs:
      - data: list of tuples. Each tuple represents a class and should contain
              the input data (a numpy array where each row is an event)
              and label (a scalar) for a class. Example: [(X, +1), (Y,-1)].
      - hiddenNodes: a list stating how many nodes to be used in each hidden layer
                    (do not specify nodes for the input and output layers).
      - trfFunction: a list of torch.nn available transfer functions for all hidden layers and the output layer.
      - testSize: the percentual size [0,1] of the testing set. Defaults to 0.5.
      - evalFunc: the evalutiaon function to use ('mse' or 'sp'). Defaults to 'sp'.
      - nTrains: how many weights initialization to perform. Defaults to 10.
      - batchSize: epochs batch size. Defaults to zero (use all available data).
      - epochs: number of training epochs. Defaults to 100.
      - numROC: ROC size, when evaluating performance. Defaults to 200.
      - nDeals: how many cross-validation deals to execute. Defaults to 20.

     Returns a populated ObjectView object created by create_results_structure
     including the results for all deals.
    """

    skf = StratifiedShuffleSplit(n_splits = nDeals, test_size = testSize)
    cvm = dm.CrossValidationManager(skf, data)
    ret = create_results_structure(nDeals, epochs, numROC)
    for i, train, test in cvm.split(returnCounter = True):
        numNodes = [train[0].shape[1]] + hiddenNodes + [train[1].shape[1]]
        net = BinaryClassificationNetwork(numNodes, trfFunction, evalFunc = evalFunc)
        trnError, tstError = net.fit(train, test, epochs, numInitializations = nTrains, batch_size = batchSize)
        ret = save_results(ret, i, trnError, tstError, net, train, test)
    return ret
